import numpy as np
from scipy.sparse import issparse
import scipy.stats
import pandas as pd
import tensorflow as tf

import NaiveDE
from anndata import AnnData

from enum import Enum, auto
from typing import Optional, Union

import logging

from .distance_cache import DistanceCache
from .kernels import Linear, SquaredExponential, Cosine


def get_dtype(df: pd.DataFrame, msg="Data frame"):
    dtys = df.dtypes.unique()
    if dtys.size > 1:
        logging.warning("%s has more than one dtype, selecting the first one" % msg)
    return dtys[0]


def normalize_counts(
    adata: AnnData,
    sizefactorcol: Optional[str] = None,
    layer: Optional[str] = None,
    copy: bool = False,
):
    if copy:
        adata = adata.copy()

    if sizefactorcol is None:
        sizefactors = pd.DataFrame({"sizefactors": calc_sizefactors(adata, layer=layer)})
        sizefactorcol = "np.log(sizefactors)"
    else:
        sizefactors = adata.obs
    X = adata.X if layer is None else adata.layers[layer]
    stabilized = NaiveDE.stabilize(dense_slice(X.T))
    regressed = np.asarray(NaiveDE.regress_out(sizefactors, stabilized, sizefactorcol).T)
    if layer is None:
        adata.X = regressed
    else:
        adata.layers[layer] = regressed
    return adata


def dense_slice(slice):
    if issparse(slice):
        slice = slice.toarray()
    else:
        slice = np.asarray(slice)  # work around anndata.ArrayView
    return np.squeeze(slice)


def bh_adjust(pvals):
    order = np.argsort(pvals)
    alpha = np.minimum(
        1, np.maximum.accumulate(len(pvals) / np.arange(1, len(pvals) + 1) * pvals[order])
    )
    return alpha[np.argsort(order)]


def calc_sizefactors(adata: AnnData, layer=None):
    X = adata.X if layer is None else adata.layers[layer]
    return np.asarray(X.sum(axis=1)).squeeze()


def get_l_limits(cache: DistanceCache):
    R2 = cache.squaredEuclideanDistance
    R2 = R2[R2 > 1e-8]

    l_min = tf.sqrt(tf.reduce_min(R2)) * 2.0
    l_max = tf.sqrt(tf.reduce_max(R2))

    return l_min, l_max


def factory(kern: str, cache: DistanceCache, lengthscale: Optional[float] = None):
    if kern == "linear":
        return Linear(cache)
    elif kern == "SE":
        return SquaredExponential(cache, lengthscale=lengthscale)
    elif kern == "PER":
        return Cosine(cache, lengthscale=lengthscale)
    else:
        raise ValueError("unknown kernel")


def kspace_walk(kernel_space: dict, cache: DistanceCache):
    for kern, lengthscales in kernel_space.items():
        try:
            for l in lengthscales:
                yield factory(kern, cache, l), kern
        except TypeError:
            yield factory(kern, cache, lengthscales), kern


def default_kernel_space(cache: DistanceCache):
    l_min, l_max = get_l_limits(cache)
    return {
        "SE": np.logspace(np.log10(l_min), np.log10(l_max), 5),
        "PER": np.logspace(np.log10(l_min), np.log10(l_max), 5),
    }


def concat_tensors(tens):
    return tf.concat([tf.reshape(t, (-1,)) for t in tens], axis=0)


def assign_concat(x, vars):
    offset = 0
    for v in vars:
        newval = tf.reshape(x[offset : (offset + tf.size(v))], v.shape)
        v.assign(newval)
        offset += tf.size(v)


def gower_factor(mat, varcomp=1):
    """Gower normalization factor for covariance matric K

    Based on https://github.com/PMBio/limix/blob/master/limix/utils/preprocess.py
    """
    return (
        varcomp
        * (tf.linalg.trace(mat) - tf.reduce_sum(tf.reduce_mean(mat, axis=0)))
        / tf.cast(tf.shape(mat)[0] - 1, mat.dtype)
    )


def quantile_normalize(mat):
    idx = np.argsort(mat, axis=0) + 0.5
    return scipy.stats.norm(loc=0, scale=1).ppf(idx / mat.shape[0])



from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from typing import Optional, Union, List, Tuple

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from gpflow import default_float
from gpflow.utilities import to_default_float

tfd = tfp.distributions
from scipy.optimize import minimize

from .kernels import Kernel

from enum import Enum, auto
import math


@dataclass(frozen=True)
class ScoreTestResults:
    kappa: Union[float, tf.Tensor]
    nu: Union[float, tf.Tensor]
    U_tilde: Union[float, tf.Tensor]
    e_tilde: Union[float, tf.Tensor]
    I_tilde: Union[float, tf.Tensor]
    pval: Union[float, tf.Tensor]

    def to_dict(self):
        ret = {}
        for f in fields(self):
            obj = getattr(self, f.name)
            if tf.is_tensor(obj):
                obj = obj.numpy()
            ret[f.name] = obj
        return ret


def combine_pvalues(
    results: Union[ScoreTestResults, List[ScoreTestResults], tf.Tensor, np.ndarray]
) -> float:
    if isinstance(results, ScoreTestResults):
        pvals = results.pval
    elif isinstance(results, list):
        pvals = tf.stack([r.pval for r in results], axis=0)
    elif tf.is_tensor(results):
        pvals = results
    elif isinstance(results, np.ndarray):
        pvals = tf.convert_to_tensor(pvals)
    else:
        raise TypeError("Unknown type for results.")

    comb = tf.reduce_mean(tf.tan((0.5 - pvals) * math.pi))
    return tfd.Cauchy(
        tf.convert_to_tensor(0, comb.dtype), tf.convert_to_tensor(1, comb.dtype)
    ).survival_function(comb)


class ScoreTest(ABC):
    @dataclass
    class NullModel(ABC):
        pass

    def __init__(
        self,
        omnibus: bool = False,
        kernel: Optional[Union[Kernel, List[Kernel]]] = None,
        yidx: Optional[tf.Tensor] = None,
    ):
        self._yidx = yidx
        self.omnibus = omnibus
        self.n = None
        if kernel is not None:
            self.kernel = kernel

    def __call__(
        self, y: tf.Tensor, nullmodel: Optional[NullModel] = None
    ) -> Tuple[ScoreTestResults, NullModel]:
        y = tf.squeeze(y)
        if self._yidx is not None:
            y = tf.gather(y, self._yidx)
        try:
            if nullmodel is None:
                nullmodel = self._fit_null(y)
            stat, e_tilde, I_tau_tau = self._test(y, nullmodel)
            return self._calc_test(stat, e_tilde, I_tau_tau), nullmodel
        except TypeError as e:
            if y.dtype is not default_float():
                raise
            # raise TypeError(
            #     f"Value vector has wrong dtype. Expected: {repr(default_float())}, given: {repr(y.dtype)}"
            # )
            else:
                raise

    @property
    def kernel(self) -> List[Kernel]:
        return self.kernel

    @kernel.setter
    def kernel(self, kernel: Union[Kernel, List[Kernel]]):
        self._kernel = [kernel] if isinstance(kernel, Kernel) else kernel
        if len(self._kernel) > 1:
            if self.omnibus:
                self._K = self._kernel[0].K()
                for k in self._kernel[1:]:
                    self._K += k.K()
            else:
                self._K = tf.stack([k.K() for k in kernel], axis=0)
        else:
            self._K = self._kernel[0].K()
        self._K = to_default_float(self._K)
        self.n = tf.shape(self._K)[0]

        if self._yidx is not None:
            x, y = tf.meshgrid(self._yidx, self._yidx)
            idx = tf.reshape(tf.stack((y, x), axis=2), (-1, 2))
            if tf.size(tf.shape(self._K)) > 2:
                bdim = tf.shape(self._K)[0]
                idx = tf.tile(idx, (bdim, 1))
                idx = tf.concat(
                    (
                        tf.repeat(
                            tf.range(bdim, dtype=self._yidx.dtype), tf.square(tf.size(self._yidx))
                        )[:, tf.newaxis],
                        idx,
                    ),
                    axis=1,
                )
            self._K = tf.squeeze(
                tf.reshape(
                    tf.gather_nd(self._K, idx),
                    (-1, tf.size(self._yidx), tf.size(self._yidx)),
                )
            )

    @abstractmethod
    def _test(
        self, y: tf.Tensor, nullmodel: NullModel
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
        pass

    @abstractmethod
    def _fit_null(self, y: tf.Tensor) -> NullModel:
        pass

    @staticmethod
    def _calc_test(stat, e_tilde, I_tau_tau) -> ScoreTestResults:
        kappa = I_tau_tau / (2 * e_tilde)
        nu = 2 * e_tilde**2 / I_tau_tau
        pval = tfd.Chi2(nu).survival_function(stat / kappa)
        return ScoreTestResults(kappa, nu, stat, e_tilde, I_tau_tau, pval)


class NegativeBinomialScoreTest(ScoreTest):
    @dataclass
    class NullModel(ScoreTest.NullModel):
        mu: tf.Tensor
        alpha: tf.Tensor

    def __init__(
        self,
        sizefactors: tf.Tensor,
        omnibus: bool = False,
        kernel: Optional[Union[Kernel, List[Kernel]]] = None,
    ):
        self.sizefactors = tf.squeeze(tf.cast(sizefactors, tf.float64))
        if tf.rank(self.sizefactors) > 1:
            raise ValueError("Size factors vector must have rank 1")

        yidx = tf.cast(tf.squeeze(tf.where(self.sizefactors > 0)), tf.int32)
        if tf.shape(yidx)[0] != tf.shape(self.sizefactors)[0]:
            self.sizefactors = tf.gather(self.sizefactors, yidx)
        else:
            yidx = None
        super().__init__(omnibus, kernel, yidx)

    def _fit_null(self, y: tf.Tensor) -> NullModel:
        scaledy = tf.cast(y, tf.float64) / self.sizefactors
        res = minimize(
            lambda *args: self._negative_negbinom_loglik(*args).numpy(),
            x0=[
                tf.math.log(tf.reduce_mean(scaledy)),
                tf.math.log(
                    tf.maximum(1e-8, self._moments_dispersion_estimate(scaledy, self.sizefactors))
                ),
            ],
            args=(tf.cast(y, tf.float64), self.sizefactors),
            jac=lambda *args: self._grad_negative_negbinom_loglik(*args).numpy(),
            method="bfgs",
        )
        mu = tf.exp(res.x[0]) * self.sizefactors
        alpha = tf.exp(res.x[1])
        return self.NullModel(mu, alpha)

    def _test(
        self, y: tf.Tensor, nullmodel: NullModel
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
        return self._do_test(
            self._K,
            to_default_float(y),
            to_default_float(nullmodel.alpha),
            to_default_float(nullmodel.mu),
        )

    @staticmethod
    @tf.function(experimental_compile=True)
    def _do_test(
        K: tf.Tensor, rawy: tf.Tensor, alpha: tf.Tensor, mu: tf.Tensor
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
        W = mu / (1 + alpha * mu)
        stat = 0.5 * tf.reduce_sum(
            (rawy / mu - 1) * W * tf.tensordot(K, W * (rawy / mu - 1), axes=(-1, -1)), axis=-1
        )

        P = tf.linalg.diag(W) - W[:, tf.newaxis] * W[tf.newaxis, :] / tf.reduce_sum(W)
        PK = W[:, tf.newaxis] * K - W[:, tf.newaxis] * ((W[tf.newaxis, :] @ K) / tf.reduce_sum(W))
        trace_PK = tf.linalg.trace(PK)
        e_tilde = 0.5 * trace_PK
        I_tau_tau = 0.5 * tf.reduce_sum(PK * PK, axis=(-2, -1))
        I_tau_theta = 0.5 * tf.reduce_sum(PK * P, axis=(-2, -1))
        I_theta_theta = 0.5 * tf.reduce_sum(tf.square(P), axis=(-2, -1))
        I_tau_tau_tilde = I_tau_tau - tf.square(I_tau_theta) / I_theta_theta

        return stat, e_tilde, I_tau_tau_tilde

    @staticmethod
    @tf.function(experimental_compile=True)
    def _moments_dispersion_estimate(y, sizefactors):
        """
        This is lifted from the first DESeq paper
        """
        v = tf.math.reduce_variance(y)
        m = tf.reduce_mean(y)
        s = tf.reduce_mean(1 / sizefactors)
        return (v - s * m) / tf.square(m)

    @staticmethod
    @tf.function(experimental_compile=True)
    def _negative_negbinom_loglik(params, counts, sizefactors):
        logmu = params[0]
        logalpha = params[1]
        mus = tf.exp(logmu) * sizefactors
        nexpalpha = tf.exp(-logalpha)
        ct_plus_alpha = counts + nexpalpha
        return -tf.reduce_sum(
            tf.math.lgamma(ct_plus_alpha)
            - tf.math.lgamma(nexpalpha)
            - ct_plus_alpha * tf.math.log(1 + tf.exp(logalpha) * mus)
            + counts * logalpha
            + counts * tf.math.log(mus)
            - tf.math.lgamma(counts + 1)
        )

    @staticmethod
    @tf.function(experimental_compile=True)
    def _grad_negative_negbinom_loglik(params, counts, sizefactors):
        logmu = params[0]
        logalpha = params[1]
        mu = tf.exp(logmu)
        mus = mu * sizefactors
        nexpalpha = tf.exp(-logalpha)
        one_alpha_mu = 1 + tf.exp(logalpha) * mus

        grad0 = tf.reduce_sum((counts - mus) / one_alpha_mu)  # d/d_mu
        grad1 = tf.reduce_sum(
            nexpalpha
            * (
                tf.math.digamma(nexpalpha)
                - tf.math.digamma(counts + nexpalpha)
                + tf.math.log(one_alpha_mu)
            )
            + (counts - mus) / one_alpha_mu
        )  # d/d_alpha
        return -tf.convert_to_tensor((grad0, grad1))



from abc import ABCMeta, abstractmethod
from typing import Union
from dataclasses import dataclass

import numpy as np
import scipy
from scipy import optimize
from scipy.misc import derivative
from scipy.stats import chi2

from .kernels import Kernel


class Model:
    def __init__(self, X: np.ndarray, kernel: Kernel):
        self.X = X
        self.n = X.shape[0]
        self.kernel = kernel

        self._K = None
        self._y = None

    def _reset(self):
        pass

    @property
    def K(self):
        if self._K is not None:
            return self._K
        else:
            return self.kernel.K(self.X)

    @property
    def n_parameters(self) -> float:
        return 0

    @property
    def FSV(self) -> float:
        return np.nan

    @property
    def s2_FSV(self) -> float:
        return np.nan

    @property
    def logdelta(self) -> float:
        return np.nan

    @logdelta.setter
    def logdelta(self, ld: float):
        pass

    @property
    def s2_logdelta(self) -> float:
        return np.nan

    @property
    def delta(self) -> float:
        return np.nan

    @delta.setter
    def delta(self, nd: float):
        pass

    @property
    def s2_delta(self) -> float:
        return np.nan

    @property
    def mu(self) -> float:
        return np.nan

    @property
    def sigma_s2(self) -> float:
        return np.nan

    @property
    def sigma_n2(self) -> float:
        return np.nan

    @property
    def y(self) -> np.ndarray:
        return self._y

    @y.setter
    def y(self, newy: np.ndarray):
        self._y = newy
        self._reset()
        self._ychanged()

    def optimize(self):
        pass

    @property
    def log_marginal_likelihood(self) -> float:
        self._check_y()
        return self._lml()

    def _lml(self):
        return np.nan

    def _ychanged(self):
        pass

    def _check_y(self):
        if self.y is None:
            raise RuntimeError("assign observed values first")
        if self.y.shape[0] != self.n:
            raise RuntimeError("different numbers of observations in y and X")

    def __enter__(self):
        self._K = self.K

    def __exit__(self, *args):
        self._K = None


class GPModel(Model):
    def __init__(self, X: np.ndarray, kernel: Kernel):
        super().__init__(X, kernel)

        self._y = None
        self._logdelta = 0
        self._reset()

    def __enter__(self):
        super().__enter__()
        self._reset()
        # Gower normalization factor for covariance matric K
        # Based on https://github.com/PMBio/limix/blob/master/limix/utils/preprocess.py
        self.gower = (np.trace(self.K) - np.sum(np.mean(self.K, axis=0))) / (self.n - 1)

        return self

    def __exit__(self, *args):
        super().__exit__(*args)

    def _reset(self):
        self._s2_FSV = None
        self._s2_delta = None
        self._s2_logdelta = None
        self._mu = None
        self._sigma_s2 = None
        self._sigma_n2 = None

    @property
    @abstractmethod
    def n_parameters(self):
        pass

    @property
    def FSV(self):
        return self.gower / (self.delta + self.gower)

    @property
    def s2_FSV(self):
        if self._s2_FSV is None:
            self._s2_FSV = self._calc_s2_FSV()
        return self._s2_FSV

    @property
    def logdelta(self) -> float:
        return self._logdelta

    @logdelta.setter
    def logdelta(self, ld: float):
        self._logdelta = ld
        self._reset()
        self._logdeltachanged()

    @property
    def s2_logdelta(self) -> float:
        if self._s2_logdelta is None:
            self._s2_logdelta = self._calc_s2_logdelta()
        return self._s2_logdelta

    @property
    def s2_delta(self) -> float:
        if self._s2_delta is None:
            self._s2_delta = self._calc_s2_delta()
        return self._s2_delta

    def _logdeltachanged(self):
        pass

    @property
    def delta(self) -> float:
        return np.exp(self.logdelta)

    @delta.setter
    def delta(self, d: float):
        self.logdelta = np.log(d)

    def _objective(self, func):
        def obj(logdelta):
            self.logdelta = logdelta
            return func()

        return obj

    def optimize(self):
        res = optimize.minimize(
            self._objective(lambda: -self.log_marginal_likelihood),
            10,
            method="l-bfgs-b",
            bounds=[(-10, 20)],
            jac=False,
            options={"eps": 1e-4},
        )
        self.logdelta = res.x[0]
        return res

    @abstractmethod
    def _lml(self):
        pass

    @property
    def mu(self) -> float:
        if self._mu is None:
            self._mu = self._calc_mu()
        return self._mu

    @property
    def sigma_s2(self) -> float:
        if self._sigma_s2 is None:
            self._sigma_s2 = self._calc_sigma_s2()
        return self._sigma_s2

    @property
    def sigma_n2(self) -> float:
        if self._sigma_n2 is None:
            self._sigma_n2 = self._calc_sigma_n2()
        return self._sigma_n2

    @abstractmethod
    def _calc_mu(self) -> float:
        pass

    @abstractmethod
    def _calc_sigma_s2(self) -> float:
        pass

    @abstractmethod
    def _calc_sigma_s2(self) -> float:
        pass

    def _calc_s2_logdelta(self) -> float:
        ld = self.logdelta
        s2 = -1 / derivative(
            self._objective(lambda: self.log_marginal_likelihood), self.logdelta, n=2
        )
        self.logdelta = ld
        return s2

    def _calc_s2_delta(self) -> float:
        return self.s2_logdelta * self.delta

    def _calc_s2_FSV(self) -> float:
        ld = self.logdelta
        der = derivative(self._objective(lambda: self.FSV), self.logdelta, n=1)
        self.logdelta = ld
        return der**2 / self.s2_logdelta


class SGPR(GPModel):
    def __init__(self, X: np.ndarray, Z: np.ndarray, kern: Kernel):
        super().__init__(X, kernel=kern)
        self.Z = Z
        self.z = Z.shape[0]

    def __enter__(self):
        super().__enter__()
        K_uu = self.kernel.K(self.Z)
        K_uf = self.kernel.K(self.Z, self.X)
        K_ff = self.kernel.K_diag(self.X)

        L = np.linalg.cholesky(K_uu + 1e-6 * np.eye(self.z))
        LK_uf = scipy.linalg.solve_triangular(L, K_uf, lower=True)
        A = LK_uf @ LK_uf.T

        self._Lambda, U = np.linalg.eigh(A)
        self._B = U.T @ LK_uf
        self._B1 = np.sum(self._B, axis=-1)
        self._By = None
        self._traceterm = np.sum(K_ff) - np.trace(A)

        return self

    def __exit__(self, *args):
        super().__exit__(*args)
        self._Lambda = None
        self._B = None
        self._B1 = None
        self._By = None
        self._traceterm = None

    @property
    def n_parameters(self) -> float:
        return 3

    def _lml(self) -> float:
        delta = self.delta
        return 0.5 * (
            -self.n * np.log(2 * np.pi)
            - self.z * np.log(self.sigma_s2)
            - (self.n - self.z) * np.log(self.sigma_n2)
            - np.sum(np.log(delta + self._Lambda))
            - self.n
            - self._traceterm / delta
        )

    def _residual_quadratic(self) -> float:
        self._check_y()
        return (
            np.sum(self.y**2)
            - 2 * np.sum(self.y) * self.mu
            + self.mu**2 * self.n
            - np.sum((self._By - self._B1 * self.mu) ** 2 / self._dL)
        )

    def _calc_mu(self):
        self._check_y()
        sy = np.sum(self.y)
        ytones = np.sum(self._By * self._B1 / self._dL)
        onesones = np.sum(self._B1**2 / self._dL)

        return (sy - ytones) / (self.n - onesones)

    def _calc_sigma_s2(self):
        return self.sigma_n2 / self.delta

    def _calc_sigma_n2(self):
        return self._residual_quadratic() / self.n

    def _ychanged(self):
        if self._B is not None:
            self._By = np.dot(self._B, self.y)

    def _logdeltachanged(self):
        self._dL = self.delta + self._Lambda


class GPR(GPModel):
    def __init__(self, X: np.ndarray, kern: Kernel):
        super().__init__(X, kernel=kern)

    def __enter__(self):
        super().__enter__()

        K = self.kernel.K(self.X)
        self._Lambda, self._U = np.linalg.eigh(K)
        self._U1 = np.sum(self._U, axis=0)
        self._Uy = None

        return self

    def __exit__(self, *args):
        super().__exit__(*args)
        self._Lambda = self._U = None
        self._U1 = None
        self._Uy = None

    @property
    def n_parameters(self) -> float:
        return 3

    def _lml(self) -> float:
        return 0.5 * (
            -self.n * np.log(2 * np.pi)
            - np.sum(np.log(self._dL))
            - self.n
            - self.n * np.log(self.sigma_s2)
        )

    def _residual_quadratic(self) -> float:
        return np.sum((self._Uy - self._U1 * self.mu) ** 2 / self._dL)

    def _calc_mu(self) -> float:
        return np.sum(self._U1 * self._Uy / self._dL) / np.sum(np.square(self._U1) / self._dL)

    def _calc_sigma_s2(self) -> float:
        return self._residual_quadratic() / self.n

    def _calc_sigma_n2(self) -> float:
        return self.sigma_s2 * self.delta

    def _ychanged(self):
        if self._U is not None:
            self._Uy = np.dot(self._U.T, self.y)

    def _logdeltachanged(self):
        self._dL = self.delta + self._Lambda


class Constant(Model):
    def __init__(self, X: np.ndarray):
        super().__init__(X, Kernel())

    def _reset(self):
        self._mu = None
        self._s2 = None

    @property
    def n_parameters(self) -> float:
        return 2

    def optimize(self):
        self._check_y()
        self._mu = np.mean(self.y)
        self._s2 = np.var(self.y, ddof=0)
        return optimize.OptimizeResult(success=True)

    @property
    def mu(self) -> float:
        if self._mu is None:
            self.optimize()
        return self._mu

    @property
    def sigma_n2(self) -> float:
        if self._s2 is None:
            self.optimize()
        return self._s2

    def _lml(self) -> float:
        return (
            -0.5 * self.n * np.log(2 * np.pi * self.sigma_n2)
            - 0.5 * np.sum(np.square(self.y - self.mu)) / self.sigma_n2
        )


class Null(Model):
    def __init__(self, X: np.ndarray):
        super().__init__(X, Kernel())

    def _reset(self):
        self._s2 = None

    @property
    def n_parameters(self) -> float:
        return 1

    def optimize(self):
        self._check_y()
        self._s2 = np.sum(np.square(self.y)) / self.n
        return optimize.OptimizeResult(success=True)

    @property
    def sigma_n2(self) -> float:
        if self._s2 is None:
            self.optimize()
        return self._s2

    def _lml(self) -> float:
        return (
            -0.5 * self.n * np.log(2 * np.pi * self.sigma_n2)
            - 0.5 * np.square(self.y) / self.sigma_n2
        )


def model_factory(X: np.ndarray, Z: Union[np.ndarray, None], kern: Kernel, *args, **kwargs):
    if Z is None:
        return GPR(X, kern, *args, **kwargs)
    else:
        return SGPR(X, Z, kern, *args, **kwargs)



import pandas as pd

def relabel_interface(idata, clust_key):
    node_labels_text = idata.uns['cell_meta'][clust_key]
    idata.obs[f'A_label_{clust_key}'] = node_labels_text.loc[idata.obs['A']].astype(str).to_numpy()
    idata.obs[f'B_label_{clust_key}'] = node_labels_text.loc[idata.obs['B']].astype(str).to_numpy()
    node_labels = idata.uns['cell_meta'][clust_key].astype('category').cat.codes
    idata.obs[f'A_label_int_{clust_key}'] = node_labels.loc[idata.obs['A']].to_numpy()
    idata.obs[f'B_label_int_{clust_key}'] = node_labels.loc[idata.obs['B']].to_numpy()
    idata.obs['label_1'] = idata.obs[f'A_label_int_{clust_key}'].astype(str) + idata.obs[f'B_label_int_{clust_key}'].astype(str)
    idata.obs['label_2'] = idata.obs[f'B_label_int_{clust_key}'].astype(str) + idata.obs[f'A_label_int_{clust_key}'].astype(str)
    idata.obs[f'{clust_key}_int'] = idata.obs[['label_1', 'label_2']].astype(int).max(axis=1).astype(str).astype('category')
    label_1 = idata.obs[f'A_label_{clust_key}'].astype(str) + '_' + idata.obs[f'B_label_{clust_key}'].astype(str).to_numpy()
    label_2 = idata.obs[f'B_label_{clust_key}'].astype(str) + '_' + idata.obs[f'A_label_{clust_key}'].astype(str).to_numpy()
    pick = idata.obs[['label_1', 'label_2']].astype(int).idxmax(axis=1).to_numpy()
    text_label = [label_1[i] if x=='label_1' else label_2[i] for i,x in enumerate(pick)]
    idata.obs[f'label_{clust_key}'] = text_label
    idata.obs[f'label_{clust_key}'] = idata.obs[f'label_{clust_key}'].astype('category')
    print(f'Added key label_{clust_key} in idata.obs')
    
def scored_spot_interface(idata):       
    belonging = {}
    cells = idata.uns['cell_meta'].index
    for i in cells:
        belonging[i] = []
    for pair in idata.obs.reset_index()[['index','A', 'B']].to_numpy():
        belonging[pair[1]].append(pair[0])
        belonging[pair[2]].append(pair[0])
    score = pd.DataFrame(idata.obsm['pattern_score'], index=idata.obs_names)
    df = pd.concat([score.loc[belonging[c]].mean() for c in cells], axis=1).T     
    df.index = cells
    idata.uns['cell_pattern'] = df
    print(f'Added key cell_pattern in idata.uns')

def interaction_spot_interface(idata):       
    belonging = {}
    cells = idata.uns['cell_meta'].index
    for i in cells:
        belonging[i] = []
    for pair in idata.obs.reset_index()[['index','A', 'B']].to_numpy():
        belonging[pair[1]].append(pair[0])
        belonging[pair[2]].append(pair[0])
    score = idata.to_df()
    df = pd.concat([score.loc[belonging[c]].mean() for c in cells], axis=1).T     
    df.index = cells
    idata.uns['cell_score'] = df
    print(f'Added key cell_score in idata.uns')

def adata_moranI(adata, out_f, n_jobs=10):
    import scanpy as sc
    import squidpy as sq
    adata_copy = adata.copy()
    sc.pp.normalize_total(adata_copy, target_sum=1e4)
    sc.pp.log1p(adata_copy)
    sc.pp.highly_variable_genes(adata_copy, flavor='seurat_v3', n_top_genes=5000)
    genes = adata_copy[:, adata_copy.var.highly_variable].var_names.values[:5000]

    sq.gr.spatial_neighbors(adata, key_added='spatial')
    sq.gr.spatial_autocorr(
        adata,
        genes=genes,
        mode="moran",
        n_perms=1000,
        n_jobs=n_jobs,
    )
    adata.uns['moranI'].to_csv(f'{out_f}svg_moranI.csv')
    
def get_marker_df(idata, logfoldchanges_threhold=1):
    marker_df = pd.concat([pd.DataFrame(idata.uns['rank_genes_groups']['pvals_adj']).melt(),
           pd.DataFrame(idata.uns['rank_genes_groups']['logfoldchanges']).melt()['value'],
           pd.DataFrame(idata.uns['rank_genes_groups']['names']).melt()['value']], axis=1)
    marker_df.columns = ['cluster', 'pvals_adj', 'logfoldchanges', 'names']
    marker_df  = marker_df[(marker_df.pvals_adj < 0.05) & (marker_df.logfoldchanges > logfoldchanges_threhold)]
    print(marker_df.cluster.value_counts())
    return marker_df

def save_runningtime(idata, save=''):
    obj = {}
    obj['nnSVG_time'] = idata.uns['nnSVG_time']
    obj['SOMDE_time'] = idata.uns['SOMDE_time']
    obj['SpatialDE_time'] = idata.uns['SpatialDE_time']
    obj['SPARKX_time'] = idata.uns['SPARKX_time']
    obj['scGCO_time'] = idata.uns['scGCO_time']
    obj['moranI_time'] = idata.uns['moranI_time']
    obj['gearyC_time'] = idata.uns['gearyC_time']
    obj['interface'] = idata.shape[0]
    obj['LRI'] = idata.shape[1]
    if save != '':
        import json
        with open(save, 'w') as f:
            json.dump(obj, f)
    return obj


from typing import List, Optional, Union

import numpy as np
from scipy.sparse import issparse
import tensorflow as tf

from anndata import AnnData


class AnnDataIterator:
    def __init__(
        self,
        adata: AnnData,
        genes: Optional[List[str]] = None,
        layer: Optional[str] = None,
        dtype: Optional[Union[np.dtype, tf.DType]] = None,
    ):
        self.adata = adata
        self.genes = genes
        if self.genes is None:
            self.genes = self.adata.var_names
        self.layer = layer
        if dtype is not None:
            outtype = dtype
        elif self.layer is None:
            outtype = self.adata.X.dtype
        else:
            outtype = self.adata.layers[layer].dtype
        self.output_types = (outtype, tf.string)

    def __call__(self):
        for i, g in enumerate(self.genes):
            if self.layer is None:
                data = self.adata.X[:, i]
            else:
                data = self.adata.layers[self.layer][:, i]
            if issparse(data):
                data = data.toarray()
            with tf.device(tf.DeviceSpec(device_type="CPU").to_string()):
                gene = tf.convert_to_tensor(g)
            yield tf.convert_to_tensor(np.squeeze(data), dtype=self.output_types[0]), gene


class AnnDataDataset(tf.data.Dataset):
    def __new__(
        cls,
        adata: AnnData,
        genes: Optional[List[str]] = None,
        layer: Optional[str] = None,
        dtype: Optional[Union[np.dtype, tf.DType]] = None,
    ):
        it = AnnDataIterator(adata, genes, layer, dtype)
        return (
            tf.data.Dataset.from_generator(it, output_types=it.output_types)
            .repeat(1)
            .prefetch(tf.data.experimental.AUTOTUNE)
        )



from collections import namedtuple
from typing import Optional, List

import numpy as np
import scipy

import tensorflow as tf
import gpflow
from gpflow import default_float
from gpflow.utilities import to_default_float

from .util import gower_factor, quantile_normalize
from .score_test import ScoreTest
from .optimizer import MultiScipyOptimizer


class SVCA(tf.Module):
    _fracvar = namedtuple("FractionVariance", "intrinsic environmental noise")
    _fracvar_interact = namedtuple("FractionVraiance", "intrinsic environmental interaction noise")

    def __init__(
        self,
        expression: np.ndarray,
        X: np.ndarray,
        sizefactors: np.ndarray,
        kernel: Optional[gpflow.kernels.Kernel] = None,
    ):
        self.expression = to_default_float(expression)
        self.sizefactors = to_default_float(sizefactors)
        self._ncells, self._ngenes = tf.shape(self.expression)
        self._X = to_default_float(X)

        self._current_expression = tf.Variable(
            tf.zeros((self._ncells, self._ngenes - 1), dtype=default_float()), trainable=False
        )
        self.intrinsic_variance_matrix = tf.Variable(
            tf.zeros((self._ncells, self._ncells), dtype=default_float()), trainable=False
        )
        self._sigmas = gpflow.Parameter(
            tf.ones((4,), dtype=default_float()), transform=gpflow.utilities.positive(lower=1e-9)
        )
        self._currentgene = tf.Variable(0, dtype=tf.int32, trainable=False)
        self.muhat = tf.Variable(tf.ones((self._ncells,), dtype=default_float()), trainable=False)

        self.kernel = kernel

        self._opt = MultiScipyOptimizer(lambda: -self.profile_log_reml(), self.trainable_variables)
        self._use_interactions = tf.Variable(False, dtype=tf.bool, trainable=False)

        self._old_interactions = False

    @property
    def sizefactors(self) -> np.ndarray:
        return self._sizefactors

    @sizefactors.setter
    def sizefactors(self, sizefactors: np.ndarray):
        self._sizefactors = np.squeeze(sizefactors)
        if len(self._sizefactors.shape) != 1:
            raise ValueError("Size factors vector must have rank 1")
        self._log_sizefactors = tf.squeeze(tf.math.log(to_default_float(sizefactors)))

    @property
    def kernel(self):
        return self._kernel

    @kernel.setter
    def kernel(self, kern):
        self._kernel = kern
        self._init_kern = gpflow.utilities.read_values(kern)

    @property
    def currentgene(self) -> int:
        return self._currentgene.numpy()

    @currentgene.setter
    def currentgene(self, gene: int):
        gene = tf.cast(gene, self._ngenes.dtype)
        if gene < 0 or gene >= self._ngenes:
            raise IndexError(f"gene must be between 0 and {self._ngenes}")

        self._currentgene.assign(gene)
        idx = [] if gene == 0 else tf.range(gene)
        idx = (
            tf.concat((idx, tf.range(gene + 1, self._ngenes)), axis=0)
            if gene < self._ngenes - 1
            else idx
        )
        self._current_expression.assign(
            tf.gather(self.expression, idx, axis=1) / self._sizefactors[:, tf.newaxis]
        )

        intvar = tf.matmul(self._current_expression, self._current_expression, transpose_b=True)
        self.intrinsic_variance_matrix.assign(intvar / gower_factor(intvar))

        muhat = self.expression[:, gene]
        muhat = tf.where(muhat < 1, 1, muhat)  # avoid problems with log link
        self.muhat.assign(muhat)

        self._sigmas.assign(
            tf.fill(
                (4,),
                0.25
                * tf.math.reduce_variance(
                    tf.math.log(self.expression[:, gene] + 1) - self._log_sizefactors
                ),
            )
        )
        if self._kernel is not None:
            gpflow.utilities.multiple_assign(self._kernel, self._init_kern)

    def profile_log_reml(self):
        Vchol = tf.linalg.cholesky(self.V())
        r = self._r(Vchol)
        quad = tf.tensordot(
            r,
            tf.squeeze(tf.linalg.cholesky_solve(Vchol, r[:, tf.newaxis])),
            axes=(-1, -1),
        )
        ldet = tf.reduce_sum(tf.math.log(tf.linalg.diag_part(Vchol)))
        ldet2 = tf.math.log(
            tf.reduce_sum(
                tf.linalg.cholesky_solve(Vchol, tf.ones((self._ncells, 1), dtype=default_float()))
            )
        )

        return -ldet - 0.5 * quad - 0.5 * ldet2

    def _alphahat(self, Vchol):
        Vinvnu = tf.linalg.cholesky_solve(Vchol, self.nu[:, tf.newaxis])
        VinvX = tf.linalg.cholesky_solve(Vchol, tf.ones((self._ncells, 1), dtype=default_float()))
        return tf.reduce_sum(Vinvnu) / tf.reduce_sum(VinvX)

    def alphahat(self):
        return self._alphahat(tf.linalg.cholesky(self.V()))

    def _betahat(self, Vchol):
        return tf.squeeze(self.D() @ tf.linalg.cholesky_solve(Vchol, self._r(Vchol)[:, tf.newaxis]))

    def betahat(self):
        return self._betahat(tf.linalg.cholesky(self.V()))

    @property
    def nu(self):
        return (
            tf.math.log(self.muhat)
            + self.expression[:, self._currentgene] / self.muhat
            - 1
            - self._log_sizefactors
        )

    def _r(self, Vchol):
        return self.nu - self._alphahat(Vchol)

    def r(self):
        Vchol = tf.linalg.cholesky(self.V())
        return self._r(Vchol)

    @tf.function
    def estimate_muhat(self):
        Vchol = tf.linalg.cholesky(self.V())
        self.muhat.assign(
            tf.exp(self._alphahat(Vchol) + self._betahat(Vchol) + self._log_sizefactors)
        )

    def V(self):
        V = self.D()
        V = tf.linalg.set_diag(V, tf.linalg.diag_part(V) + 1 / self.muhat)
        return V

    # no property here, apparently tf.function has a problem with conditionals in properties
    def D(self):
        var = self.intrinsic_variance + self.environmental_variance
        var = tf.linalg.set_diag(var, tf.linalg.diag_part(var) + self.noise_variance)
        if self._use_interactions:
            var += self.interaction_variance
        return var

    def dV_dsigma(self):
        if self._use_interactions:
            return tf.stack(
                (
                    self.intrinsic_variance_matrix,
                    self.environmental_variance_matrix,
                    self.interaction_variance_matrix,
                    tf.eye(self._ncells, dtype=default_float()),
                ),
                axis=0,
            )
        else:
            return tf.stack(
                (
                    self.intrinsic_variance_matrix,
                    self.environmental_variance_matrix,
                    tf.eye(self._ncells, dtype=default_float()),
                ),
                axis=0,
            )

    def fraction_variance(self):
        intrinsic = gower_factor(self.intrinsic_variance)
        environ = gower_factor(self.environmental_variance)
        noise = self.noise_variance

        totalgower = intrinsic + environ + noise
        if self._use_interactions:
            interact = gower_factor(self.interaction_variance)
            totalgower += interact

            return self._fracvar_interact(
                (intrinsic / totalgower).numpy(),
                (environ / totalgower).numpy(),
                (interact / totalgower).numpy(),
                (noise / totalgower).numpy(),
            )
        else:
            return self._fracvar(
                (intrinsic / totalgower).numpy(),
                (environ / totalgower).numpy(),
                (noise / totalgower).numpy(),
            )

    @property
    def environmental_variance_matrix(self):
        return self.kernel.K(self._X)

    @property
    def interaction_variance_matrix(self):
        envmat = self.environmental_variance_matrix
        intmat = envmat @ tf.matmul(self.intrinsic_variance_matrix, envmat, transpose_b=True)
        return intmat / gower_factor(intmat)

    @property
    def intrinsic_variance(self):
        return self._sigmas[0] * self.intrinsic_variance_matrix

    @property
    def environmental_variance(self):
        return self._sigmas[1] * self.environmental_variance_matrix

    @property
    def interaction_variance(self):
        return self._sigmas[2] * self.interaction_variance_matrix

    @property
    def noise_variance(self):
        return self._sigmas[3]

    def use_interactions(self, interact: bool):
        self._old_interactions = self._use_interactions.numpy()
        self._use_interactions.assign(interact)
        return self

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._use_interactions.assign(self._old_interactions)

    def optimize(self, abstol: float = 1e-5, maxiter: int = 1000):
        oldsigmas = self._sigmas.numpy()
        for i in range(maxiter):
            self._opt.minimize()
            sigmas = self._sigmas.numpy()
            if np.all(np.abs(sigmas - oldsigmas) < abstol):
                break
            oldsigmas = sigmas
            self.estimate_muhat()


class SVCAInteractionScoreTest(ScoreTest):
    def __init__(
        self,
        expression_matrix: np.ndarray,
        X: np.ndarray,
        sizefactors: np.ndarray,
        kernel: Optional[gpflow.kernels.Kernel] = None,
    ):
        super().__init__()
        self._model = SVCA(expression_matrix, X, sizefactors, kernel)
        self._model.use_interactions(False)

    @property
    def kernel(self) -> List[gpflow.kernels.Kernel]:
        if self._model.kernel is not None:
            return [self._model.kernel]
        else:
            return []

    @kernel.setter
    def kernel(self, kernel: gpflow.kernels.Kernel):
        self._model.kernel = kernel

    def _fit_null(self, y):
        self._model.currentgene = y
        self._model.optimize()
        return None

    def _test(self, y, nullmodel: None):
        return self._do_test(
            self._model.r(),
            self._model.V(),
            self._model.dV_dsigma(),
            self._model.interaction_variance_matrix,
        )

    @staticmethod
    @tf.function(experimental_compile=True)
    def _do_test(residual, V, dV, interaction_mat):
        cholV = tf.linalg.cholesky(V)
        Vinvres = tf.squeeze(tf.linalg.cholesky_solve(cholV, residual[:, tf.newaxis]))
        stat = 0.5 * tf.tensordot(
            Vinvres, tf.tensordot(interaction_mat, Vinvres, axes=(-1, -1)), axes=(-1, -1)
        )

        Vinv_int = tf.linalg.cholesky_solve(cholV, interaction_mat)
        Vinv_dV = tf.linalg.cholesky_solve(cholV[tf.newaxis, ...], dV)

        Vinv_X = tf.squeeze(
            tf.linalg.cholesky_solve(
                cholV, tf.ones((tf.shape(residual)[0], 1), dtype=default_float())
            )
        )
        hatMat = Vinv_X[:, tf.newaxis] * Vinv_X[tf.newaxis, :] / tf.reduce_sum(Vinv_X)

        P_int = Vinv_int - hatMat @ interaction_mat
        P_dV = Vinv_dV - hatMat[tf.newaxis, ...] @ dV

        e_tilde = 0.5 * tf.linalg.trace(P_int)
        I_tau_tau = tf.reduce_sum(tf.transpose(P_int) * P_int)
        I_tau_theta = tf.reduce_sum(tf.transpose(P_int) * P_dV, axis=[-2, -1])
        I_theta_theta = tf.reduce_sum(
            tf.linalg.matrix_transpose(P_dV[tf.newaxis, ...]) * P_dV[:, tf.newaxis, ...],
            axis=[-2, -1],
        )

        I_tau_tau_tilde = 0.5 * (
            I_tau_tau
            - tf.tensordot(
                I_tau_theta,
                tf.squeeze(tf.linalg.solve(I_theta_theta, I_tau_theta[..., tf.newaxis])),
                axes=(-1, -1),
            )
        )

        return stat, e_tilde, I_tau_tau_tilde



from typing import Optional, List, Union, Tuple
import warnings
from dataclasses import dataclass
from collections import namedtuple
from collections.abc import Iterable
from numbers import Real, Integral

import numpy as np
import tensorflow as tf

from gpflow import default_float, default_jitter, Parameter, set_trainable
from gpflow.utilities import to_default_float, positive
from gpflow.optimizers import Scipy

from anndata import AnnData

from ._internal.kernels import SquaredExponential
from ._internal.util import normalize_counts, get_l_limits
from ._internal.distance_cache import DistanceCache
from ._internal.util_mixture import prune_components, prune_labels


@dataclass(frozen=True)
class SpatialPatternParameters:
    """
    Parameters for automated expession histology.

    Args:
        nclasses: Maximum number of regions to consider. Defaults to the square root of the number of observations.
        lengthscales: List of kernel lenthscales. Defaults to a single lengthscale of the minimum distance between
            observations.
        pattern_prune_threshold: Probability threshold at which unused patterns are removed. Defaults to ``1e-6``.
        method: Optimization algorithm, must be known to ``scipy.optimize.minimize``. Defaults to ``l-bfgs-b``.
        tol: Convergence tolerance. Defaults to 1e-9.
        maxiter: Maximum number of iterations. Defaults to ``1000``.
        gamma_1: Parameter of the noise variance prior, defaults to ``1e-14``.
        gamma_2: Parameter of the noise variance prior, defaults to ``1e-14``.
        eta_1: Parameter of the Dirichlet process hyperprior, defaults to ``1``.
        eta_2: Parameter of the Dirichlet process hyperprior, defaults to ``1``.
    """

    nclasses: Optional[Integral] = None
    lengthscales: Optional[Union[Real, List[Real]]] = None
    pattern_prune_threshold: float = 1e-6
    method: str = "l-bfgs-b"
    tol: Optional[Real] = 1e-9
    maxiter: Integral = 1000
    gamma_1: Real = 1e-14
    gamma_2: Real = 1e-14
    eta_1: Real = 1
    eta_2: Real = 1

    def __post_init__(self):
        if self.nclasses is not None:
            assert not isinstance(
                self.lengthscales, Iterable
            ), "You must specify either nclasses or a list of lengthscales"
        if isinstance(self.lengthscales, Real):
            assert self.lengthscales > 0, "Lengthscales must be positive"
        elif self.lengthscales is not None:
            for l in self.lengthscales:
                assert l > 0, "Lengthscales must be positive"
        assert (
            self.pattern_prune_threshold >= 0 and self.pattern_prune_threshold <= 1
        ), "Class pruning threshold must be between 0 and 1"
        assert self.method in ("l-bfgs-b", "bfgs"), "Method must be either bfgs or l-bfgs-b"
        if self.tol is not None:
            assert self.tol > 0, "Tolerance must be greater than 0"
        assert self.maxiter >= 1, "Maximum number of iterations must greater than or equal to 1"
        assert self.gamma_1 > 0, "Gamma1 hyperparameter must be positive"
        assert self.gamma_2 > 0, "Gamma2 hyperparameter must be positive"
        assert self.eta_1 > 0, "Eta1 hyperparameter must be positive"
        assert self.eta_2 > 0, "Eta2 hyperparameter must be positive"


@dataclass(frozen=True)
class SpatialPatterns:
    """
    Results of automated expression histology.

    Args:
        converged: Whether the optimization converged.
        status: Status of the optimization.
        labels: The estimated region labels.
        pattern_probabilities: N_obs x N_patterns array with the estimated region probabilities for each observation.
        niter: Number of iterations for the optimization.
        elbo_trace: ELBO values at each iteration.
    """

    converged: bool
    status: str
    labels: np.ndarray
    pattern_probabilities: np.ndarray
    patterns: np.ndarray
    niter: int
    elbo_trace: np.ndarray


class _SpatialPatterns(tf.Module):
    def __init__(
        self,
        X: np.ndarray,
        counts: np.ndarray,
        nclasses: Integral,
        lengthscales: List[Real],
        gamma_1: Real,
        gamma_2: Real,
        eta_1: Real,
        eta_2: Real,
        rng: np.random.Generator = np.random.default_rng(),
    ):
        self.X = to_default_float(X)
        self.counts = to_default_float(counts)
        self.nsamples, self.ngenes = tf.shape(counts)
        self._fnsamples, self._fngenes = to_default_float(tf.shape(counts))
        self.nclasses = nclasses
        self._fnclasses = to_default_float(self.nclasses)

        self.gamma_1 = to_default_float(gamma_1)
        self.gamma_2 = to_default_float(gamma_2)
        self.eta_1 = to_default_float(eta_1)
        self.eta_2 = to_default_float(eta_2)

        dcache = DistanceCache(X)
        if lengthscales is None:
            l_min, l_max = get_l_limits(dcache)
            lengthscales = [0.5 * l_min]
        elif not isinstance(lengthscales, Iterable):
            lengthscales = [lengthscales]
        self.kernels = []
        Kernel = namedtuple("DecomposedKernel", "Lambda U")

        lcounts = np.unique(lengthscales, return_counts=True)
        for l, c in zip(*lcounts):
            k = SquaredExponential(dcache, lengthscale=l).K()
            S, U = tf.linalg.eigh(k)
            self.kernels.extend([Kernel(S, U)] * c)
        if len(self.kernels) == 1:
            self.kernels = self.kernels * nclasses

        self.phi = Parameter(
            rng.uniform(-0.01, 0.01, (self.ngenes, self.nclasses)), dtype=default_float()
        )  # to break ties, otherwise all gradients are the same
        self.etahat_2 = Parameter(
            eta_2 + default_jitter(), dtype=default_float(), transform=positive(lower=eta_2)
        )
        self.gammahat_2 = Parameter(
            self.gammahat_1,
            dtype=default_float(),
            transform=positive(lower=gamma_2),
        )

    @property
    def etahat_1(self):
        return self.eta_1 + self._fnclasses - 1

    @property
    def pihat(self):
        return tf.nn.softmax(self.phi, axis=1)

    @property
    def gammahat_1(self):
        return self.gamma_1 + 0.5 * self._fnsamples * self._fngenes

    @property
    def _sigmahat(self):
        return self.gammahat_1 / self.gammahat_2

    def Sigma_hat_inv(self, c=0):
        k = self.kernels[c]
        EigUt = k.U * ((self._sigmahat * self._N(c) * k.Lambda + 1) / k.Lambda)[tf.newaxis, :]
        return tf.matmul(k.U, EigUt, transpose_b=True)

    def Sigma_hat(self, c=0):
        k = self.kernels[c]
        EigUt = k.U * (k.Lambda / (self._sigmahat * self._N(c) * k.Lambda + 1))[tf.newaxis, :]
        return tf.matmul(k.U, EigUt, transpose_b=True)

    @property
    def mu_hat(self):
        ybar = self._ybar()
        return tf.stack(
            [self._mu_hat(c, ybar=ybar[:, c]) for c in tf.range(self.nclasses)],
            axis=1,
        )

    def _mu_hat(self, c=None, Sigma_hat=None, ybar=None):
        assert c is not None or Sigma_hat is not None and ybar is not None
        if Sigma_hat is None:
            Sigma_hat = self.Sigma_hat(c)
        if ybar is None:
            ybar = self._ybar(c)
        return self._sigmahat * tf.tensordot(Sigma_hat, ybar, axes=(-1, -1))

    def _ybar(self, c=None):
        if c is None:
            return self.counts @ self.pihat
        else:
            return tf.tensordot(self.counts, self.pihat[:, c], axes=(-1, -1))

    def _N(self, c=None):
        if c is None:
            return tf.reduce_sum(self.pihat, axis=0)
        else:
            return tf.reduce_sum(self.pihat[:, c])

    @property
    def _lhat(self):
        return tf.math.log(self.gammahat_2) - tf.math.digamma(self.gammahat_1)

    @property
    def _alphahat(self):
        return self.etahat_1 / self.etahat_2

    @property
    def _alphahat1(self):
        return 1 + self._N()[:-1]

    @property
    def _alphahat2(self):
        pihat_cumsum = tf.cumsum(tf.reduce_sum(self.pihat, axis=0), reverse=True)
        return pihat_cumsum[1:] + self._alphahat

    @property
    def _vhat2(self):
        return tf.math.digamma(self._alphahat1) - tf.math.digamma(self._alphahat1 + self._alphahat2)

    @property
    def _vhat3(self):
        return tf.math.digamma(self._alphahat2) - tf.math.digamma(self._alphahat1 + self._alphahat2)

    def elbo(self):
        pihat = self.pihat
        dotcounts = tf.reduce_sum(tf.square(self.counts), axis=0)
        sigmahat = self._sigmahat
        ybar = self._ybar()
        lhat = self._lhat
        N = self._N()

        term1 = 0.5 * (self._fnsamples * self._fngenes * lhat + sigmahat * tf.reduce_sum(dotcounts))

        term2 = 0
        for c in range(self.nclasses):
            k = self.kernels[c]
            UTybar = tf.tensordot(k.U, ybar[:, c], axes=(0, 0))
            Lambdahat = sigmahat * N[c] * k.Lambda + 1

            ybar_mu = tf.square(sigmahat) * tf.tensordot(
                UTybar, k.Lambda / Lambdahat * UTybar, axes=(-1, -1)
            )
            pimu = N[c] * (
                sigmahat**3
                * tf.tensordot(UTybar * tf.square(k.Lambda / Lambdahat), UTybar, axes=(-1, -1))
                + tf.reduce_sum(k.Lambda / Lambdahat)
            )
            muinv_ybar = tf.square(sigmahat) * tf.tensordot(
                UTybar * k.Lambda / tf.square(Lambdahat), UTybar, axes=(-1, -1)
            )
            trace_inv = tf.reduce_sum(1 / Lambdahat)
            logdet = tf.reduce_sum(tf.math.log(Lambdahat))

            term2 += ybar_mu - 0.5 * (pimu + muinv_ybar + trace_inv + logdet)

        term3 = tf.reduce_sum(N[:-1] * self._vhat2)

        vhat3_cumsum = tf.concat(((0,), tf.cumsum(self._vhat3)), axis=0)
        term4 = tf.reduce_sum(N * vhat3_cumsum)

        term5 = tf.reduce_sum((self._alphahat - self._alphahat2) * self._vhat3)
        term6 = tf.reduce_sum(pihat * self.phi)

        term7 = tf.reduce_sum(tf.reduce_logsumexp(self.phi, axis=1))
        term8 = (
            -self.gammahat_1 * (1 - tf.math.log(self.gammahat_2))
            + tf.math.lgamma(self.gammahat_1)
            + (self.gammahat_1 + 1) * lhat
        )
        term9 = tf.reduce_sum((self._alphahat1 - 1) * self._vhat2)
        term10 = tf.reduce_sum(tf.math.lbeta(tf.stack((self._alphahat1, self._alphahat2), axis=1)))
        term11 = (
            -(self.eta_1 + self._fnclasses) * tf.math.log(self.etahat_2)
            + (self.etahat_2 - self.eta_2) * self._alphahat
        )

        elbo = (
            -term1 + term2 + term3 + term4 + term5 - term6 + term7 + term8 - term9 + term10 + term11
        ) / (self._fnsamples * self._fnclasses * self._fngenes)
        return elbo


def spatial_patterns(
    adata: AnnData,
    genes: Optional[List[str]] = None,
    normalized=False,
    spatial_key="spatial",
    layer: Optional[str] = None,
    params: SpatialPatternParameters = SpatialPatternParameters(),
    rng: np.random.Generator = np.random.default_rng(),
    copy: bool = False,
) -> Tuple[SpatialPatterns, Union[AnnData, None]]:
    """
    Detect spatial patterns of gene expression and assign genes to patterns.

    Uses a Gaussian process mixture. A Dirichlet process prior allows
    to automatically determine the number of distinct regions in the dataset.

    Args:
        adata: The annotated data matrix.
        genes: List of genes to base the analysis on. Defaults to all genes.
        normalized: Whether the data are already normalized to an approximately Gaussian likelihood.
            If ``False``, they will be normalized using the workflow from Svensson et al, 2018.
        spatial_key: Key in ``adata.obsm`` where the spatial coordinates are stored.
        layer: Name of the AnnData object layer to use. By default ``adata.X`` is used.
        params: Parameters for the algorithm, e.g. prior distributions, spatial smoothness, etc.
        rng: Random number generator.
        copy: Whether to return a copy of ``adata`` with results or write the results into ``adata``
            in-place.

    Returns:
        A tuple. The first element is a :py:class:`SpatialPatterns`, the second is ``None`` if ``copy == False``
        or an ``AnnData`` object. Patterns will be in ``adata.obs["spatial_pattern_0"]``, ...,
        ``adata.obs["spatial_pattern_n"]``.
    """
    if not normalized and genes is None:
        warnings.warn(
            "normalized is False and no genes are given. Assuming that adata contains complete data set, will normalize and fit a GP for every gene."
        )
    data = normalize_counts(adata, copy=True) if not normalized else adata
    if genes is not None:
        data = data[:, genes]

    X = data.obsm[spatial_key]
    counts = data.X if layer is None else adata.layers[layer]

    # This is important, we only care about co-expression, not absolute levels.
    counts = counts - tf.reduce_mean(counts, axis=0)
    counts = counts / tf.math.reduce_std(counts, axis=0)

    nclasses = params.nclasses
    if nclasses is None:
        if isinstance(params.lengthscales, Iterable):
            nclasses = len(params.lengthscales)
        else:
            nclasses = int(np.ceil(np.sqrt(data.n_vars)))

    patterns = _SpatialPatterns(
        X,
        counts,
        nclasses,
        params.lengthscales,
        params.gamma_1,
        params.gamma_2,
        params.eta_1,
        params.eta_2,
        rng,
    )
    opt = Scipy()
    elbo_trace = [patterns.elbo()]
    res = opt.minimize(
        lambda: -patterns.elbo(),
        patterns.trainable_variables,
        method=params.method,
        step_callback=lambda step, vars, vals: elbo_trace.append(patterns.elbo()),
        tol=params.tol,
        options={"maxiter": params.maxiter},
    )

    prune_threshold = tf.convert_to_tensor(params.pattern_prune_threshold, dtype=default_float())
    idx, labels = prune_components(
        tf.argmax(patterns.pihat, axis=1),
        tf.transpose(patterns.pihat),
        prune_threshold,
        everything=True,
    )
    print('normalizing')
    pihat = tf.linalg.normalize(tf.gather(patterns.pihat, idx, axis=1), ord=1, axis=1)[0]
    patterns = tf.gather(patterns.mu_hat, idx, axis=1).numpy()

    if copy:
        adata = adata.copy()
        toreturn = adata
    else:
        toreturn = None
    for i in range(patterns.shape[1]):
        adata.obs[f"spatial_pattern_{i}"] = patterns[:, i]

    return (
        SpatialPatterns(
            converged=res.success,
            status=res.message,
            labels=labels.numpy(),
            pattern_probabilities=pihat.numpy(),
            patterns=patterns,
            niter=res.nit,
            elbo_trace=np.asarray(elbo_trace),
        ),
        toreturn,
    )



from typing import Union, Optional
from abc import ABCMeta, abstractmethod
import math

import tensorflow as tf
from gpflow import default_float
from gpflow.utilities import to_default_float
from gpflow.utilities.ops import square_distance, difference_matrix

from .distance_cache import DistanceCache


def scale(X: tf.Tensor, lengthscale: Optional[float] = 1):
    if X is not None:
        return X / lengthscale
    else:
        return X


def scaled_difference_matrix(
    X: tf.Tensor,
    Y: Optional[tf.Tensor] = None,
    lengthscale: Optional[float] = 1,
):
    return difference_matrix(scale(X, lengthscale), scale(Y, lengthscale))


def scaled_squared_distance(
    X: tf.Tensor, Y: Optional[tf.Tensor] = None, lengthscale: Optional[float] = 1
):
    return square_distance(scale(X, lengthscale), scale(Y, lengthscale))


class Kernel(metaclass=ABCMeta):
    def __init__(self, cache: DistanceCache):
        self._cache = cache

    def K(self, X: Optional[tf.Tensor] = None, Y: Optional[tf.Tensor] = None):
        if X is None and Y is not None:
            X, Y = Y, X
        if (X is None or X is self._cache.X) and Y is None:
            return self._K(cache=True)
        else:
            X = to_default_float(X)
            if Y is not None:
                Y = to_default_float(Y)
        return self._K(X, Y)

    def K_diag(self, X: Optional[tf.Tensor]):
        if X is None:
            return self._K_diag(cache=True)
        else:
            return self._K_diag(to_default_float(X))

    @abstractmethod
    def _K(self, X: Optional[tf.Tensor] = None, Y: Optional[tf.Tensor] = None, cache: bool = False):
        pass

    @abstractmethod
    def _K_diag(self, X: Optional[tf.Tensor] = None, cache: bool = False):
        pass


class StationaryKernel(Kernel):
    def __init__(self, cache: DistanceCache, lengthscale=1):
        super().__init__(cache)
        self.lengthscale = lengthscale

    def _K_diag(self, X: Optional[tf.Tensor] = None, cache: bool = False):
        if cache:
            n = self._cache.X.shape[0]
        else:
            n = X.shape[0]
        return self._K_diag_impl(n)

    def _K_diag_impl(self, n: int):
        return tf.repeat(tf.convert_to_tensor(1, dtype=default_float()), n)


class SquaredExponential(StationaryKernel):
    def _K(self, X: Optional[tf.Tensor] = None, Y: Optional[tf.Tensor] = None, cache: bool = False):
        if cache:
            dist = self._cache.squaredEuclideanDistance / self.lengthscale**2
        else:
            dist = scaled_squared_distance(X, Y, self.lengthscale)
        return tf.exp(-0.5 * dist)


class Cosine(StationaryKernel):
    def _K(self, X: Optional[tf.Tensor] = None, Y: Optional[tf.Tensor] = None, cache: bool = False):
        if cache:
            dist = self._cache.sumOfDifferences / self.lengthscale
        else:
            dist = tf.reduce_sum(scaled_difference_matrix(X, Y, self.lengthscale), axis=-1)
        return tf.cos(2 * math.pi * dist)


class Linear(Kernel):
    def _K(self, X: tf.Tensor, Y: Optional[tf.Tensor] = None):
        if Y is None:
            Y = X
        return tf.sum(X[:, tf.newaxis, :] * Y[tf.newaxis, ...], axis=-1)

    def _K_diag(self, X: tf.Tensor, cache: bool = False):
        if cache:
            X = self._cache.X
        return tf.sum(tf.square(X), axis=-1)



from typing import Optional

import tensorflow as tf


@tf.function(experimental_relax_shapes=True)
def prune_components(labels: tf.Tensor, pihat: tf.Tensor, threshold: tf.Tensor, everything=False):
    toretain = tf.squeeze(tf.where(tf.reduce_any(pihat > threshold, axis=1)), axis=1)
    if not everything:
        toretain = tf.range(
            tf.reduce_max(toretain) + 1
        )  # we can not prune everything during optimization, then vhat3_cumsum would be wrong
        return tf.cast(toretain, labels.dtype), labels
    return prune_labels(labels, tf.cast(toretain, labels.dtype))


@tf.function(experimental_relax_shapes=True)
def prune_labels(labels: tf.Tensor, toretain: Optional[tf.Tensor] = None):
    if toretain is None:
        ulabels, _ = tf.unique(labels)
        toretain = tf.sort(ulabels)
    else:
        toretain = tf.sort(toretain)
    diffs = toretain[1:] - toretain[:-1]
    missing = tf.cast(tf.where(diffs > 1), labels.dtype)
    if tf.size(missing) > 0:
        missing = tf.squeeze(missing, axis=1)
        todrop = tf.TensorArray(labels.dtype, size=tf.size(missing), infer_shape=False)
        shift = tf.cast(0, labels.dtype)
        for i in tf.range(tf.size(missing)):
            m = missing[i]
            idx = tf.where(labels > toretain[m] - shift)
            shift += diffs[m] - 1
            labels = tf.tensor_scatter_nd_sub(labels, idx, tf.repeat(diffs[m] - 1, tf.size(idx)))
            todrop = todrop.write(i, tf.range(toretain[m] + 1, toretain[m] + diffs[m]))
        todrop = todrop.concat()
        if toretain[0] > 0:
            todrop = tf.concat((tf.range(toretain[0]), todrop), axis=0)
            labels = labels - toretain[0]
        idx = tf.squeeze(
            tf.sparse.to_dense(
                tf.sets.difference(
                    tf.range(tf.reduce_max(toretain) + 1)[tf.newaxis, :],
                    tf.cast(todrop[tf.newaxis, :], dtype=labels.dtype),
                )
            )
        )
    else:
        idx = tf.cast(tf.range(tf.size(toretain)), labels.dtype)
    return idx, labels
